/*
 * *************************************************************************
 *
 * Trimble Navigation, Ltd.
 * OEM Products Development Group
 * P.O. Box 3642
 * 645 North Mary Avenue
 * Sunnyvale, California 94088-3642
 *
 * Corporate Headquarter:
 *    Telephone:  (408) 481-8000
 *    Fax:        (408) 481-6005
 *
 * Technical Support Center:
 *    Telephone:  (800) 767-4822	(U.S. and Canada)
 *                (408) 481-6940    (outside U.S. and Canada)
 *    Fax:        (408) 481-6020
 *    BBS:        (408) 481-7800
 *    e-mail:     trimble_support@trimble.com
 *		ftp://ftp.trimble.com/pub/sct/embedded/bin
 *
 * *************************************************************************
 *
 * TSIP_UTL.C contains five types of functions:
 *
 * 1) Borland-compatible window control functions:  The Borland version of
 * TSIPCHAT supports two windows: a command and report window and an auto
 * window.  The command and report window displays the command list, command
 * entries and parameters, and display of requested packets.  The auto window
 * scrolls through unrequested reports, which are generated according to the
 * I/O options specified.
 *
 * The following are Borland C++ functions used to create text windows.  If
 * suitable functions can be found in the Microsoft Visual C++ Compiler, the
 * windowing feaures may be recreated. (Creating TSIPCHAT without BORLAND
 * defined results in all commands, both auto and requested, scrolling up the
 * screen.)
 *
 *    ;; create a window
 *    void window (short left, short top, short right, short bottom)
 *
 *    ;; set backgrounf and foreground colors
 *    void textattr (char attr)
 *
 *    ;; go to row x, colw y w.r.t window origin
 *    void gotoxy (short x, short y)
 *
 *    ;; report row relative to window origin
 *    short wherex (void)
 *
 *    ;; report column relative to window origin
 *    short wherey (void)
 *
 *    ;; clear window to the end of the line
 *    void clrscr (void), clreol (void)
 *
 *    ;; get text attributes
 *    void gettextinfo (struct *textinfo)
 *
 * 2) User interface functions:  Specifically, these are functions that query
 * integer, double, or YES/NO inputs.  There is also a function for toggling
 * file storage on and off.
 *
 * 3) Time functions:  These functions support receiver time initialization.
 *
 * 4) Storage file functions:  These functions open and close binary
 * storage files.
 *
 * 5) Serial port functions:  These functions set serial port
 * parameters.  They are DOS-specific.
 *
 * *************************************************************************
 *
 */

#define END_STRING '\0'
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "t_struct.h"
#include "tsip_sio.h"
#include "tsipincl.h"

/**/
/*  *******************  1. CONSOLE WINDOW CONTROL ****************** */
/*
 * The following routines set up two windows: one which prints only
 * commands and requested reports, the other only unrequested reports.
 * The routines are set up for Borland compilers.  Similar functions can
 * be found with Microsoft compilers.  Generic versions are included
 * for non-Borland compilers.
 */

/* backspace, enter, printable characters except control characters */
#include <conio.h>

/* text attributes for different screen functions */
static int
	auto_y,
	text_color[3];
static struct text_info
	t_info,
	w_info[2];
#define UPCURSOR ('H'<<8)
#define DOWNCURSOR ('P'<<8)
#define LEFTCURSOR ('K'<<8)
#define RIGHTCURSOR ('M'<<8)

/* current window = command or requested reply, or auto report */
/* TRUE if current window is command window */
static int
	curr_win = AUTOWIN;

void show_crlf (void)
{
	clreol ();
	cprintf ("\r\n");
	clreol ();
}

int which_window (void)
{
	return (curr_win);
}

void switch_window (int new_win)
{
	struct text_info *old_info, *new_info;

	old_info = &w_info[curr_win];
	gettextinfo (old_info);
	if (new_win == curr_win) {
		return;
	}
	curr_win = new_win;
	new_info = &w_info[curr_win];

	switch (new_win) {
	case AUTOWIN:
		/* auto window is the bottom of the screen; unrequested reports */
		/* are printed here in scrolled mode */
		window (1, old_info->cury + 1, 80, 24);

		textattr (text_color[new_win]);
		/* start at bottom of auto window and scroll */
		gotoxy (80, auto_y - (old_info->cury + 1));
		break;

	case CMDWIN:
		/* cmd window is the top of the screen; commands and requested  */
		/* reports are printed in scrolled mode in blue or reverse text */

		/* autowin absolute line coordinate */
		auto_y = old_info->cury + old_info->wintop;

		window (1, 1, 80, 24);
		textattr (text_color[new_win]);
		gotoxy (new_info->curx, new_info->cury);
		clreol ();
	  break;
	}
}

void new_cmd_window (void)
{
	/* blank out old command window */
	if (curr_win == AUTOWIN) {
		gettextinfo (&w_info[curr_win]);
		auto_y = w_info[curr_win].cury + w_info[curr_win].wintop;
	}
	textattr (text_color[AUTOWIN]);
	if (w_info[CMDWIN].cury > 1) {
		window (1, 2, 80, w_info[CMDWIN].cury);
		clrscr ();
	}
	w_info[CMDWIN].cury = 1;
	w_info[CMDWIN].curx = 1;

	/* perch in the auto window at the last location */
	curr_win = AUTOWIN;
	window (1, 2, 80, 24);
	gotoxy (80, auto_y - 2);

	switch_window (CMDWIN);
	clreol ();
	highvideo ();
	textbackground (BANNER_BACKGROUND);
	textcolor (BANNER_TEXT);
	cprintf (
		"       TSIPCHAT %s(%s)    Use <Esc> to Exit; Use '?' for Help Screen",
		TSIPCHAT_VERNUM, TSIP_VERNUM);
	clreol ();
	normvideo ();
	textcolor (LIGHTGRAY);
	textbackground (BLUE);
	gotoxy (1, 1);
}

void clr_cmd_window (int mode)
{
	int
		i;
	static int
		endline;
	switch_window (CMDWIN);
	if (mode) {
		for (i = w_info[CMDWIN].cury + 1; i <= endline; i++) {
			gotoxy (1, i);
			clreol ();
		}
	} else {
		endline = w_info[CMDWIN].cury;
		for (i = 1; i <= endline; i++) {
			gotoxy (1,i);
			clreol ();
		}
		gotoxy (1, 1);
		highvideo ();
		textbackground (BANNER_BACKGROUND);
		textcolor (BANNER_TEXT);
		cprintf (
			"            TSIPCHAT %s:%s     Use <CR> to Return to Normal Mode",
			TSIPCHAT_VERNUM, TSIP_VERNUM);
		clreol (); show_crlf ();
		normvideo ();
		textcolor (LIGHTGRAY);
		textbackground (BLUE);
	}
	switch_window (AUTOWIN);
}

void start_menu_window (void)
{
	textattr (text_color[MENUWIN]);
}

void end_menu_window (void)
{
	textattr (text_color[CMDWIN]);
}

void initialize_screen (void)
{
	/* get initial text attributes and save for restore before exit. */
	gettextinfo (&t_info);
	clrscr();
	text_color[AUTOWIN] = LIGHTGRAY + (BLACK << 4);
	text_color[CMDWIN]  = LIGHTGRAY + (BLUE  << 4);
	text_color[MENUWIN] = LIGHTGRAY + (BLUE  << 4);
	w_info[CMDWIN].cury = 1;
	w_info[CMDWIN].curx = 1;

	/* start in auto window */
	textattr (text_color[AUTOWIN]);
	clrscr ();
	gotoxy (1, 25);
}

void reset_screen (void)
{
	window (1,1,80,25);
	/* return text to original */
	textattr (t_info.normattr);
	gotoxy (80, 25);
	show_crlf ();
   clrscr();
}

/**/
/* ****************** 2. USER INPUT PROMPTS ********************* */
/*
 * Prompt functions: allows user to supply a character string in response
 * to a prompt.
 */

/* These routines are to set cmdesc (TRUE = abort command).
 * Used in T_COMMND.C.
 */
static int cmd_abort = FALSE;

int cmd_esc (void)
{
   return cmd_abort;
}

void set_cmd_esc (int x)
{
   cmd_abort = x;
}

/* Common prompt escape sequence. */
static int esc_out (int c)
{
	int
   	cmdesc,
      cchar;

	/* backspace, <CR>, printable characters except control characters */
   cchar = (c & 0xFF);
   if (!cchar)
   {
   	cmdesc = (c!=UPCURSOR) && (c!=DOWNCURSOR) &&
      	(c!=LEFTCURSOR) && (c!=RIGHTCURSOR);
   }
   else if (cchar == 0x08)
   {
   	cmdesc = FALSE;	/* back space is OK */
   }
	else if (cchar == 0x0D)
   {
   	cmdesc = FALSE;	/* CR is OK */
   }
   else if (isprint(c))
   {
   	if (iscntrl(c))
      {
      	cmdesc = TRUE;
      }
      else
      {
      	cmdesc = FALSE;
      }
   }
   else
   {
   	cmdesc = TRUE;	/* unprintable */
   }
	set_cmd_esc (cmdesc);
	return cmdesc;
}

static int get_prompt_string (char kbbuf[], short maxi)
{
	int kbch;
	int i;

	for (i = 0; i < maxi; ) {
		kbch = ProcessInputBytesWhileWaitingForKBHit ();
		if (kbch == 0x08) {		      /* backspace */
			cprintf ("%c", kbch);
			if (i) i--;
		}
      else if (kbch == 0x0D)
      {    /* <CR> */
			break;
		}
      else if (esc_out(kbch))
      {
			i = 0;
			break;
		}
      else if (isprint(kbch))
      {
			cprintf ("%c", kbch);
			kbbuf[i] = (char)kbch;
			i++;
		}
	}
	/* terminate string */
	kbbuf[i] = END_STRING;
	show_crlf();
	set_cmd_esc (!i);
	return i;
}

/* prompt for an integer from user; return long integer */
long int ask_lint (char *prompt)
{
	long int	n;
	char kbbuf[30];

   n = 0;
	clreol ();
	cprintf (prompt);
	if (get_prompt_string (kbbuf, 30))
   {
		set_cmd_esc (sscanf (kbbuf, "%ld", &n) < 1);
	}
	return n;
}

/* prompt for an integer from user; return integer */
int ask_int (char *prompt)
{
	return (int)ask_lint(prompt);
}

/* prompt for an integer from user; return unsigned char */
unsigned char ask_byte (char *prompt)
{
	return (unsigned char)ask_lint(prompt);
}

/* prompt for a choice from among sixteen or fewer */
/* Routine returns numerical value corresponding to choice. */
unsigned char PickFromMenu (char *MenuTitle, char *MenuItems[], int nopts)
{
	int
   	iopt, newopt,
      ichoice, newichoice, numchoices,
      colwide, pad,
      numcols, irow,
      numrows, jcol,
      strlenx, maxstrlen,
      yMenuTitle, xEndMenuTitle, yBottomMenu,
      order[100];
   int
   	kbch;

	set_cmd_esc (FALSE);

   /*
   ** nopts is sizeof(MenuItems)
   ** to compute number of MenuItems, must divide by sizeof(char*)
   */
   nopts /= (int)sizeof(char*);

   /* compute number of actual choices, ignoring void choices */
   for (iopt=0, maxstrlen=0, numchoices=0; iopt<nopts; iopt++)
   {
   	if (MenuItems[iopt][0] != END_STRING)
  	   {
	   	strlenx = strlen(MenuItems[iopt]);
   	   if (strlenx > maxstrlen) maxstrlen = strlenx;
         order[numchoices++] = iopt;
      }
   }

   /*
   ** write menu title on first line.
   ** save locations of top row & end of prompt for later use.
   */
	cprintf ("%s", MenuTitle);
   xEndMenuTitle = wherex();
   yMenuTitle = wherey();
  	clreol();
   gotoxy (1, yMenuTitle+1);
   cprintf (
"________________________________CHOICES____________________________________");

   /* menu items go below title line n rows & columns */
   /* compute column size */
   if (numchoices > 4)
   {
	   numcols = 78 / (maxstrlen + 6);
   	if (numcols > numchoices) numcols = numchoices;
	   numrows = (numchoices - 1) / numcols + 1;
      if (numrows < numcols) numrows = (int)sqrt(numchoices-1)+1;
   	numcols = (numchoices - 1) / numrows + 1;

   }
   else
   {
   	numrows = numchoices;
      numcols = 1;
   }
	colwide = 78 / numcols;
   pad = (colwide - maxstrlen)/2+2;
   if (pad > 10) pad = 10;


   /* write menu items */
   for (irow = 0; irow < numrows; irow++)
   {
		int ij;
	  	show_crlf();
     	yBottomMenu = wherey();
      cprintf ("**");
   	for (jcol = 0; jcol < numcols; jcol++)
      {
      	ichoice = irow + jcol*numrows;
         if (ichoice >= numchoices) break;
        	gotoxy (jcol*colwide + pad, yBottomMenu);
      	ij = order[ichoice];
  			cprintf ("%X = %s", ij, MenuItems[ij]);
      }
   }

   /*
   ** highlight current choice among menu items & wait for keyboard input.
   ** hex character will make a direct choice of menu item.
   ** space bar cycles among the menu items.
   */
   yBottomMenu = wherey();
   for (ichoice=0, newopt = -1; (newopt<0) || (newopt>nopts); )
   {
      iopt = order[ichoice];
      /* highlight current choice in menu items */
      irow = ichoice%numrows;
      jcol = ichoice/numrows;
      gotoxy (jcol*colwide + pad, yMenuTitle+2+irow);
      highvideo();
      cprintf ("%X = %s", iopt, MenuItems[iopt]);
      /* enter current choice at end of title line  */
  	  	gotoxy (xEndMenuTitle, yMenuTitle);
      cprintf ("%X = %s", iopt, MenuItems[iopt]);
	  	clreol();
      lowvideo();
      /* wait at bottom; prevents AUTOWIN from overwriting CMDWIN area */
      gotoxy (80, yBottomMenu);
		kbch = ProcessInputBytesWhileWaitingForKBHit ();

		if (esc_out(kbch))
      {
			return 0;
		}

      switch (kbch)
      {
      case '0':case '1':case '2':case '3':case '4':
      case '5':case '6':case '7':case '8':case '9':
      	/* direct choice 1-9 */
	   	newopt = kbch - '0';
         break;
      case 'a':case 'b':case 'c':case 'd':case 'e':case 'f':
      	/* direct choice  10-15 */
      	newopt = kbch - 'a' + 10;
         break;
      case 'A':case 'B':case 'C':case 'D':case 'E':case 'F':
      	/* direct choice 10-15 */
      	newopt = kbch - 'A' + 10;
         break;
      case UPCURSOR:
	      /* de-highlight old choice in menu; bump up the index by one */
        	gotoxy (jcol*colwide + pad, yMenuTitle+2+irow);
  			cprintf ("%X = %s", iopt, MenuItems[iopt]);
         if (irow>0) irow--;
      	ichoice = irow + jcol*numrows;
         break;
      case DOWNCURSOR:
	      /* de-highlight old choice in menu; bump up the index by one */
        	gotoxy (jcol*colwide + pad, yMenuTitle+2+irow);
  			cprintf ("%X = %s", iopt, MenuItems[iopt]);
         irow++;
         if (irow>=numrows) irow = numrows-1;
      	newichoice = irow + jcol*numrows;
         if (newichoice < numchoices)
         {
         	ichoice = newichoice;
         }
         break;
      case LEFTCURSOR:
	      /* de-highlight old choice in menu; bump up the index by one */
        	gotoxy (jcol*colwide + pad, yMenuTitle+2+irow);
  			cprintf ("%X = %s", iopt, MenuItems[iopt]);
         if (jcol>0) jcol--;
      	ichoice = irow + jcol*numrows;
         break;
      case RIGHTCURSOR:
	      /* de-highlight old choice in menu; bump up the index by one */
        	gotoxy (jcol*colwide + pad, yMenuTitle+2+irow);
  			cprintf ("%X = %s", iopt, MenuItems[iopt]);
         jcol++;
         if (jcol>=numcols) jcol = numcols-1;
      	newichoice = irow + jcol*numrows;
         if (newichoice < numchoices)
         {
         	ichoice = newichoice;
         }
         break;
      case ' ':
	      /* de-highlight old choice in menu; bump up the index by one */
        	gotoxy (jcol*colwide + pad, yMenuTitle+2+irow);
  			cprintf ("%X = %s", iopt, MenuItems[iopt]);
         ichoice = (ichoice+1)%numchoices;
         break;
      case 0x0D:
      	/* final choice */
      	newopt = iopt;
         break;
      }
   }
   /* clear menu item area with AUTOWIN */
	textattr (text_color[AUTOWIN]);
	window (1, yMenuTitle+1, 80, yBottomMenu);
	clrscr ();
   /* return to CMDWIN area, write final choice, return choice */
  	textattr (text_color[CMDWIN]);
	window (1, 1, 80, 24);
   gotoxy(xEndMenuTitle, yMenuTitle);
   clreol();
	cprintf ("   %s", MenuItems[newopt]);
   show_crlf();
   return (unsigned char)newopt;
}

/* prompt user to enter an double */
double ask_dbl (char *prompt)
{
	double d = 0.0;
	char kbbuf[30];

	clreol ();
	cprintf (prompt);
	if (get_prompt_string (kbbuf, 30)) {
		set_cmd_esc (sscanf (kbbuf, "%lf", &d) < 1);
	}
	return d;
}

/* ask user for up to 20 hex codes */
unsigned char ask_hex_string (char *prompt, unsigned char *h, unsigned char maxchar)
{
	char kbbuf [90];
	unsigned char hh [30];
	int i, n;

	clreol ();
	cprintf (prompt);
   show_crlf();
	if (get_prompt_string (kbbuf, 90))
	{
		n=sscanf(kbbuf,
"%2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x",
			&hh[ 0],&hh[ 1],&hh[ 2],&hh[ 3],&hh[ 4],
         &hh[ 5],&hh[ 6],&hh[ 7],&hh[ 8],&hh[ 9],
			&hh[10],&hh[11],&hh[12],&hh[13],&hh[14],
         &hh[15],&hh[16],&hh[17],&hh[18],&hh[19]);
		if (n > 0)
		{
			if (n > maxchar) n = maxchar;
			for (i=0; i<n; i++)
			{
				h[i] = hh[i];
			}
			return (unsigned char)n;
		}
		set_cmd_esc (TRUE);
	}
	return 0;
}


int ask_char_string (char *prompt, char *h)
{
	int i;

	clreol ();
	cprintf (prompt);
	if (get_prompt_string (h, 60)) {
		for (i=0; i<60; i++)
		{
			if (h[i] == 0) break;
		}
		if (i>0) return i;
		set_cmd_esc (TRUE);
	}
	return 0;
}


/* prompt user to verify choice with a 'Y' or 'y' */
int ask_verify (void)
{
	int kbch;
	int affirm;

	cprintf (" (Y/N) ");
	kbch = ProcessInputBytesWhileWaitingForKBHit ();
	if (esc_out(kbch)) return FALSE;
	affirm = ((kbch == 'Y') || (kbch == 'y'));
	cprintf (affirm ? "YES" : "NO");
	show_crlf ();
	return affirm;
}

/**/
/* ********************  3. TIMING ROUTINES *********************** */

/* logical indicating desire to set time */
static double
	timezone_hrs = -24.0;		/* timezone offset, hours; nonsense value to start */
static int
	timezone_set = FALSE;
const int days_of_month [12] = {
		  0,  31,  59,  90, 120, 151,
		181, 212, 243, 273, 304, 334};

void GPS_time_from_PC (double *gpstime)
{
	double
		gpsday;
	short
		day, month, year,
		hour, minute, second,
      TZFlag;
   char *timezone_text[5] = {
"Before setting the clock, the timezone offset must be entered.  If the PC is",
"configured for local time, time must be adjusted to GPS (GMT) time with the",
"proper timezone offset, equal to the number of timezones east of Greenwich.",
"Timezones west of Greenwich are negative.  Add +1 for Daylight time.",
"(Value must be between -14 and 14)."};

	switch_window (CMDWIN);
   GetYMDHMS (&year, &month, &day, &hour, &minute, &second, &TZFlag);
	clreol (); show_crlf ();
	/* asctime() inserts a \n and \0 after the first 24 characters */
	cprintf ("Current PC time is %02d:%02d:%02d on %02d/%02d/%04d (DMY)",
   	hour, minute, second, day, month, year);
   show_crlf();
	clreol ();

	if (TZFlag != GMTTIME)
   {
   	if (timezone_set)
      {
			clreol (); show_crlf ();
			cprintf ("Local offset relative to GMT is %.1f hours", timezone_hrs);
			show_crlf ();
		}
		else
		{
			show_crlf ();
			cprintf (timezone_text[0]); show_crlf ();
			cprintf (timezone_text[1]); show_crlf ();
			cprintf (timezone_text[2]); show_crlf ();
			cprintf (timezone_text[3]); show_crlf ();
			cprintf (timezone_text[4]); show_crlf ();
			show_crlf ();
			timezone_set = TRUE;
			do {
				/*
				** Some time zones are not integer hour offsets: Newfoundland,
				** Saudi Arabia, India, Singapore, parts of Australia, etc.
				*/
				timezone_hrs = ask_dbl (
      	   	"Enter timezone (PST=-8, PDT=-7, EST=-5, GMT=0, JST=9): ");
				if (cmd_esc ())
            {
					timezone_set = FALSE;
					timezone_hrs = -24.;
					cprintf ("Timezone set failure");
					return;
				}
			} while ((timezone_hrs < -14.) || (timezone_hrs > 14.));
			show_crlf ();
      }
	   GetYMDHMS (&year, &month, &day, &hour, &minute, &second, &TZFlag);
	}
   else
   {
   	timezone_hrs = 0;
   }

	gpsday = day + days_of_month[month - 1] + (year - 1980) * 365 - 6;
	/* leap years - add one for every fourth previous February */
	gpsday += (int)(12*(year-1976)+month-3)/48;

	*gpstime = (second + 60. * (minute + 60. * (hour + 24. * gpsday)))
   	- (timezone_hrs * 3600.);
}

/**/
/* *******************  4. STORAGE FILE CONTROL ****************** */
/* Following two routines called by open_, close_file_storage () below.
 * Return TRUE or FALSE.
 */
static int append_or_overwrite (char *newname, int *append_mode)
/* returns need_filename */
{
	FILE *stream_ptr;

	if (newname[0] == END_STRING) return TRUE;

	stream_ptr = fopen (newname, "r");
	if (stream_ptr != NULL) {
		fclose (stream_ptr);
		cprintf ("Append to existing file '%s'?", newname);
		if (ask_verify ()) {
			*append_mode = TRUE;
			return FALSE;
		} else {
			*append_mode = FALSE;
			cprintf ("Overwrite existing file '%s'?", newname);
			return (!ask_verify ());
		}
	}
	return FALSE;
}

void open_file_storage (FILE **stream_ptr, char *filename)
{
	int
		append_mode,
		need_filename;
	char
		newname[80];
	char
		open_mode[3] = {END_STRING, END_STRING, END_STRING};

	set_cmd_esc (FALSE);
	if (*stream_ptr != NULL) return;

	append_mode = FALSE;
	strcpy (newname, filename);
	if (newname[0] == END_STRING)
   {
		need_filename = TRUE;
		cprintf ("No previous storage file.");
        clreol ();
		show_crlf ();
	}
   else
   {
		need_filename = append_or_overwrite (newname, &append_mode);
	}
	if (cmd_esc ()) return;

	do {  /* open file */
		while (need_filename) {
			cprintf ("Enter new output file name ('^Z' to abort): ");
			clreol ();
			if (!get_prompt_string (newname, 70)) return;
			/* open to read; check to see if file exists already */
			need_filename = append_or_overwrite (newname, &append_mode);
			if (cmd_esc ()) return;
		}
		(append_mode) ?(open_mode[0] = 'a'):(open_mode[0] = 'w');
		open_mode[1] = 'b';
		*stream_ptr = fopen (newname, open_mode);
		/* open file for binary write or append */
		if (*stream_ptr != NULL)
		{
			break;
		}
		/* file open failed; loop and ask for new file name. */
		cprintf ("Could not open '%s'.", newname);
		show_crlf ();
		need_filename = TRUE;
	} while (*stream_ptr == NULL);

	strcpy (filename, newname);
	cprintf ("Saving to '%s'.", filename);
	show_crlf ();
}

void close_file_storage (FILE **stream_ptr, char *filename)
{
	set_cmd_esc (FALSE);
	if (*stream_ptr == NULL) return;

	cprintf ("Currently storing data to '%s'. Close file?", filename);
	if (!ask_verify ())
   {
		show_crlf ();
		set_cmd_esc (TRUE);
		return ;
	}
	if (fclose (*stream_ptr) == EOF)
   {
	   cprintf ("File close error, '%s'", filename);
	}
   else
   {
		*stream_ptr = NULL;
		cprintf ("Closing '%s'.", filename);
	}
	show_crlf ();
}

/**/
/* *****************  5. SERIAL PORT CONTROL  ****************** */
int set_buf_serial_port (PORT_DEF *pdef)
{
	int
   	IsNMEA, IsTSIP,
   	need_params;
	unsigned int
		tempbaud, baud, databits,
      parity, stopbits, iparity;
   static char
   	*parity_text[3] = {"none", "odd", "even"};


	set_cmd_esc (FALSE);

   /* first time through - set to COM1 and initialize to 9600 8-odd-1 */
	need_params = TRUE;
	cprintf ("Change port parameters?");
	clreol ();
	if (!ask_verify ()) {
		return TRUE;
	}
   IsNMEA =
   	pdef->baud == 4800
   	&& pdef->databits == 8
      && pdef->parity == 'n'
      && pdef->stopbits == 1;
	if (need_params && !IsNMEA)
	{
		cprintf ("Set to 4800 8-none-1?");
		if (ask_verify ())
		{
	      pdef->baud = 4800;
   	   pdef->databits = 8;
      	pdef->parity = 'n';
	      pdef->stopbits = 1;
			need_params = FALSE;
		}
      if (cmd_esc()) return TRUE;
	}
   IsTSIP =
   	pdef->baud == 9600
   	&& pdef->databits == 8
      && pdef->parity == 'o'
      && pdef->stopbits == 1;
	if (need_params && !IsTSIP)
	{
		cprintf ("Set to 9600 8-odd-1?");
		if (ask_verify ())
		{
	      pdef->baud = 9600;
   	   pdef->databits = 8;
      	pdef->parity = 'o';
	      pdef->stopbits = 1;
			need_params = FALSE;
		}
      if (cmd_esc()) return TRUE;
	}
	if (need_params)
	{
		set_cmd_esc (FALSE);
      show_crlf ();
      baud = pdef->baud;
      databits = pdef->databits;
      parity = pdef->parity;
      stopbits = pdef->stopbits;

		cprintf ("Current serial I\\O setting: %d, %d-%c-%d",
			baud, databits, parity, stopbits);
		show_crlf ();

		do {
      	baud = (unsigned int)ask_lint ("Baud rate: ");
	      tempbaud = baud/75;
   	   while (!(tempbaud&1)) tempbaud >>= 1;
			if (cmd_esc ())
	      {
				show_crlf ();
				cprintf ("-- aborted --");
	         return TRUE;
			}
      } while (tempbaud!=1);

      do {
			databits = ask_int ("Data bits (7 or 8): ");
			if (cmd_esc ())
      	{
				show_crlf ();
				cprintf ("-- aborted --");
   	      return TRUE;
         }
		} while (databits<7 || databits>8);

		iparity = PickFromMenu ("Parity: ", parity_text, sizeof(parity_text));
		if (cmd_esc ())
      {
			show_crlf ();
			cprintf ("-- aborted --");
         return TRUE;
		}
      parity = parity_text[iparity][0];

      do
      {
			stopbits = ask_int ("Stop bits (1 or 2): ");
			if (cmd_esc ())
   	   {
				show_crlf ();
				cprintf ("-- aborted --");
	         return TRUE;
			}
      } while (stopbits<1 || stopbits>2);

		cprintf ("Set port to %d,  %d-%c-%d",
			baud, databits, parity, stopbits);

		if (!ask_verify ()) {
			show_crlf ();
			cprintf ("-- aborted --");
         return TRUE;
		}
      pdef->baud = baud;
      pdef->databits = databits;
      pdef->parity = parity;
      pdef->stopbits = stopbits;
	}
  	return FALSE;
}

