/*
 * *************************************************************************
 *
 * Trimble Navigation, Ltd.
 * OEM Products Development Group
 * P.O. Box 3642
 * 645 North Mary Avenue
 * Sunnyvale, California 94088-3642
 *
 * Corporate Headquarter:
 *    Telephone:  (408) 481-8000
 *    Fax:        (408) 481-6005
 *
 * Technical Support Center:
 *    Telephone:  (800) 767-4822	(U.S. and Canada)
 *                (408) 481-6940    (outside U.S. and Canada)
 *    Fax:        (408) 481-6020
 *    BBS:        (408) 481-7800
 *    e-mail:     trimble_support@trimble.com
 *		ftp://ftp.trimble.com/pub/sct/embedded/bin
 *
 * *************************************************************************
 *
 * T_REPORT.C consists of a primary function TranslateTSIPReportToText()
 * called by main().
 *
 * This function takes a character buffer that has been received as a report
 * from a TSIP device and interprets it.  The character buffer has been
 * assembled using tsip_input_proc() in T_PARSER.C.
 *
 * A large case statement directs processing to one of many mid-level
 * functions.  The mid-level functions specific to the current report
 * code passes the report buffer to the appropriate report decoder
 * rpt_0x?? () in T_PARSER.C, which converts the byte stream in rpt.buf
 * to data values approporaite for use.
 *
 * *************************************************************************
 *
 */

#include <stdio.h>
#include <math.h>
#include "t_struct.h"
#include "t_parser.h"
#include "tsipincl.h"
#include "t_labels.h"

#define GOOD_PARSE 0
#define BADID_PARSE 1
#define BADLEN_PARSE 2
#define BADDATA_PARSE 3

/* pbuf is the pointer to the current location of the text output */
static char
	*pbuf;

/* keep track of whether the message has been successfully parsed */
static short
	parsed;


/* convert time of week into day-hour-minute-second and print */
char* show_time (float time_of_week)
{
	short	days, hours, minutes;
	float seconds;
	double tow;
   static char timestring [80];

	if (time_of_week == -1.0)
   {
		sprintf(timestring, "   <No time yet>   ");
	}
	else if ((time_of_week >= 604800.0) || (time_of_week < 0.0))
   {
		sprintf(timestring, "     <Bad time>     ");
	}
   else
   {
		if (time_of_week < 604799.9) tow = time_of_week + .00000001;
		seconds = (float)fmod(tow, 60.);
		minutes =  (short) fmod(tow/60., 60.);
		hours = (short)fmod(tow / 3600., 24.);
		days = (short)(tow / 86400.0);
		sprintf(timestring, " %s %02d:%02d:%05.2f   ",
  	   	dayname[days], hours, minutes, seconds);
   }
   return timestring;
}

/**/
/* 0x3D */
static void rpt_chan_A_config (TSIPPKT *rpt)
{
	unsigned char
		tx_baud_index, rx_baud_index,
		char_format_index, stop_bits,
      tx_mode_index, rx_mode_index,
      databits, parity;
	int
		i, nbaud;

	/* unload rptbuf */
	if (rpt_0x3D (rpt,
		&tx_baud_index, &rx_baud_index, &char_format_index,
		&stop_bits, &tx_mode_index, &rx_mode_index)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nChannel A Configuration");

   nbaud = sizeof(old_baudnum);

	for (i = 0; i < nbaud; ++i) if (tx_baud_index == old_baudnum[i]) break;
	pbuf += sprintf(pbuf, "\n   Transmit speed: %s at %s",
		old_output_ch[tx_mode_index], st_baud_text_app[i]);

	for (i = 0; i < nbaud; ++i) if (rx_baud_index == old_baudnum[i]) break;
	pbuf += sprintf(pbuf, "\n   Receive speed: %s at %s",
		old_input_ch[rx_mode_index], st_baud_text_app[i]);

	databits = (unsigned char)((char_format_index & 0x03) + 5);

	parity = (unsigned char)(char_format_index >> 2);
	if (parity > 4) parity = 2;

	pbuf += sprintf(pbuf, "\n   Character format (bits/char, parity, stop bits): %d-%s-%d",
		databits, old_parity_text[parity], stop_bits);
}

/**/
/* 0x40 */
static void rpt_almanac_data_page (TSIPPKT *rpt)
{
	unsigned char
		sv_prn;
	short
		week_num;
	float
		t_zc,
		eccentricity,
		t_oa,
		i_0,
		OMEGA_dot,
		sqrt_A,
		OMEGA_0,
		omega,
		M_0;

	/* unload rptbuf */
	if (rpt_0x40 (rpt,
		&sv_prn, &week_num, &t_zc, &eccentricity, &t_oa,
		&i_0, &OMEGA_dot, &sqrt_A, &OMEGA_0, &omega, &M_0)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nAlmanac for SV %02d", sv_prn);
	pbuf += sprintf(pbuf, "\n       Captured:%15.0f %s",
   	t_zc, show_time (t_zc));
	pbuf += sprintf(pbuf, "\n           week:%15d", week_num);
	pbuf += sprintf(pbuf, "\n   Eccentricity:%15g", eccentricity);
	pbuf += sprintf(pbuf, "\n           T_oa:%15.0f %s",
   	t_oa, show_time (t_oa));
	pbuf += sprintf(pbuf, "\n            i 0:%15g", i_0);
	pbuf += sprintf(pbuf, "\n      OMEGA dot:%15g", OMEGA_dot);
	pbuf += sprintf(pbuf, "\n         sqrt A:%15g", sqrt_A);
	pbuf += sprintf(pbuf, "\n        OMEGA 0:%15g", OMEGA_0);
	pbuf += sprintf(pbuf, "\n          omega:%15g", omega);
	pbuf += sprintf(pbuf, "\n            M 0:%15g", M_0);
}

/* 0x41 */
static void rpt_GPS_time (TSIPPKT *rpt)
{
	float
		time_of_week, UTC_offset;
	short
		week_num;

	/* unload rptbuf */
	if (rpt_0x41 (rpt, &time_of_week, &UTC_offset, &week_num)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nGPS time:%s GPS week: %d   UTC offset %.1f",
   	show_time(time_of_week), week_num, UTC_offset);

}

/* 0x42 */
static void rpt_single_ECEF_position (TSIPPKT *rpt)
{
	float
		ECEF_pos[3], time_of_fix;

	/* unload rptbuf */
	if (rpt_0x42 (rpt, ECEF_pos, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nSXYZ:  %15.0f  %15.0f  %15.0f    %s",
		ECEF_pos[0], ECEF_pos[1], ECEF_pos[2],
      show_time(time_of_fix));
}

/* 0x43 */
static void rpt_single_ECEF_velocity (TSIPPKT *rpt)
{

	float
		ECEF_vel[3], freq_offset, time_of_fix;

	/* unload rptbuf */
	if (rpt_0x43 (rpt, ECEF_vel, &freq_offset, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nVelECEF: %11.3f  %11.3f  %11.3f  %12.3f%s",
		ECEF_vel[0], ECEF_vel[1], ECEF_vel[2], freq_offset,
      show_time(time_of_fix));
}

/*  0x45  */
static void rpt_SW_version (TSIPPKT *rpt) {
	unsigned char
		major_nav_version, minor_nav_version,
		nav_day, nav_month, nav_year,
		major_dsp_version, minor_dsp_version,
		dsp_day, dsp_month, dsp_year;

	/* unload rptbuf */
	if (rpt_0x45 (rpt,
		&major_nav_version, &minor_nav_version,
		&nav_day, &nav_month, &nav_year,
		&major_dsp_version, &minor_dsp_version,
		&dsp_day, &dsp_month, &dsp_year)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf,
"\nFW Versions:  Nav Proc %2d.%02d  %2d/%2d/%2d  Sig Proc %2d.%02d  %2d/%2d/%2d",
		major_nav_version, minor_nav_version, nav_day, nav_month, nav_year,
		major_dsp_version, minor_dsp_version, dsp_day, dsp_month, dsp_year);
}

/* 0x46 */
static void rpt_rcvr_health (TSIPPKT *rpt)
{
	unsigned char
		status1, status2;
	static char
		*sc_text[] = {
			"Doing position fixes",
			"Don't have GPS time yet",
			"Waiting for almanac collection",
			"DOP too high          ",
			"No satellites available",
			"Only 1 satellite available",
			"Only 2 satellites available",
			"Only 3 satellites available",
			"No satellites usable   ",
			"Only 1 satellite usable",
			"Only 2 satellites usable",
			"Only 3 satellites usable",
			"Chosen satellite unusable"};


	/* unload rptbuf */
	if (rpt_0x46 (rpt, &status1, &status2))
	{
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nRcvr status1: %s (%02Xh); ",
     	sc_text[rpt->buf[0]], status1);

	pbuf += sprintf(pbuf, "status2: %s, %s (%02Xh)",
		(status2 & 0x01)?"No BBRAM":"BBRAM OK",
		(status2 & 0x10)?"No Ant":"Ant OK",
      status2);
}

/* 0x47 */
static void rpt_SNR_all_SVs (TSIPPKT *rpt)
{
	unsigned char
		nsvs, sv_prn[12];
	short
		isv;
	float
		snr[12];

	/* unload rptbuf */
	if (rpt_0x47 (rpt, &nsvs, sv_prn, snr))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nSNR for satellites: %d", nsvs);
	for (isv = 0; isv < nsvs; isv++)
   {
		pbuf += sprintf(pbuf, "\n    SV %02d   %6.2f",
      	sv_prn[isv], snr[isv]);
	}
}

/* 0x48 */
static void rpt_GPS_system_message (TSIPPKT *rpt)
{
	unsigned char
		message[23];

	/* unload rptbuf */
	if (rpt_0x48 (rpt, message))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nGPS message: %s", message);
}

/* 0x49 */
static void rpt_almanac_health_page (TSIPPKT *rpt)
{
	short
		iprn;
	unsigned char
		sv_health [32];

	/* unload rptbuf */
	if (rpt_0x49 (rpt, sv_health))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nAlmanac health page:");
	for (iprn = 0; iprn < 32; iprn++)
   {
		if (!(iprn%5)) *pbuf++ = '\n';
		pbuf += sprintf(pbuf, "    SV%02d  %2X",
      	(iprn+1) , sv_health[iprn]);
	}
}

/* 0x4A */
static void rpt_single_lla_position (TSIPPKT *rpt) {
	short
		lat_deg, lon_deg;
	float
		lat, lon,
		alt, clock_bias, time_of_fix;
	double lat_min, lon_min;
	unsigned char
		north_south, east_west;

	if (rpt_0x4A (rpt,
		&lat, &lon, &alt, &clock_bias, &time_of_fix))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	/* convert from radians to degrees */
	lat *= (float)R2D;
	north_south = 'N';
	if (lat < 0.0)
   {
		north_south = 'S';
		lat = -lat;
	}
	lat_deg = (short)lat;
	lat_min = (lat - lat_deg) * 60.0;

	lon *= (float)R2D;
	east_west = 'E';
	if (lon < 0.0)
   {
		east_west = 'W';
		lon = -lon;
	}
	lon_deg = (short)lon;
	lon_min = (lon - lon_deg) * 60.0;

	pbuf += sprintf(pbuf, "\nSLLA: %4d: %06.3f  %c%5d:%06.3f  %c%10.2f  %12.2f%s",
		lat_deg, lat_min, north_south,
		lon_deg, lon_min, east_west,
		alt, clock_bias,
		show_time(time_of_fix));
}

/* 0x4A */
static void rpt_ref_alt (TSIPPKT *rpt) {

	float
		alt, dummy;
	unsigned char
		alt_flag;

	if (rpt_0x4A_2 (rpt,
		&alt, &dummy, &alt_flag))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nReference Alt:   %.1f m;    %s",
   	alt, alt_flag?"ON":"OFF");
}

/* 0x4B */
static void rpt_rcvr_id_and_status (TSIPPKT *rpt)
{

	unsigned char
		machine_id, status3, status4;

	/* unload rptbuf */
	if (rpt_0x4B (rpt, &machine_id, &status3, &status4))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nRcvr Machine ID: %d; Status3 = %s, %s (%02Xh)",
   	machine_id,
		(status3 & 0x02)?"No RTC":"RTC OK",
		(status3 & 0x08)?"No Alm":"Alm OK",
		status3);
}

/* 0x4C */
static void rpt_operating_parameters (TSIPPKT *rpt)
{
	unsigned char
		dyn_code;
	float
		el_mask, snr_mask, dop_mask, dop_switch;

	/* unload rptbuf */
	if (rpt_0x4C (rpt, &dyn_code, &el_mask,
		&snr_mask, &dop_mask, &dop_switch))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nOperating Parameters:");
	pbuf += sprintf(pbuf, "\n     Dynamics code = %d %s",
   	dyn_code, dyn_text[dyn_code]);
	pbuf += sprintf(pbuf, "\n     Elevation mask = %.2f", el_mask * R2D);
	pbuf += sprintf(pbuf, "\n     SNR mask = %.2f", snr_mask);
	pbuf += sprintf(pbuf, "\n     DOP mask = %.2f", dop_mask);
	pbuf += sprintf(pbuf, "\n     DOP switch = %.2f", dop_switch);
}

/* 0x4D */
static void rpt_oscillator_offset (TSIPPKT *rpt)
{
	float
		osc_offset;

	/* unload rptbuf */
	if (rpt_0x4D (rpt, &osc_offset))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nOscillator offset: %.2f Hz = %.3f PPM",
   	osc_offset, osc_offset/1575.42);
}

/* 0x4E */
static void rpt_GPS_time_set_response (TSIPPKT *rpt)
{

	unsigned char
		response;

	/* unload rptbuf */
	if (rpt_0x4E (rpt, &response))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	switch (response)
   {
	case 'Y':
		pbuf += sprintf(pbuf, "\nTime set accepted");
		break;

	case 'N':
		pbuf += sprintf(pbuf, "\nTime set rejected or not required");
		break;

	default:
		parsed = BADDATA_PARSE;
	}
}

/* 0x4F */
static void rpt_UTC_offset (TSIPPKT *rpt)
{
	double
		a0;
	float
		a1, time_of_data;
	short
		dt_ls, wn_t, wn_lsf, dn, dt_lsf;

	/* unload rptbuf */
	if (rpt_0x4F (rpt, &a0, &a1, &time_of_data,
		&dt_ls, &wn_t, &wn_lsf, &dn, &dt_lsf)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nUTC Correction Data");
	pbuf += sprintf(pbuf, "\n   A_0         = %g  ", a0);
	pbuf += sprintf(pbuf, "\n   A_1         = %g  ", a1);
	pbuf += sprintf(pbuf, "\n   delta_t_LS  = %d  ", dt_ls);
	pbuf += sprintf(pbuf, "\n   t_ot        = %.0f  ", time_of_data);
	pbuf += sprintf(pbuf, "\n   WN_t        = %d  ", wn_t );
	pbuf += sprintf(pbuf, "\n   WN_LSF      = %d  ", wn_lsf );
	pbuf += sprintf(pbuf, "\n   DN          = %d  ", dn );
	pbuf += sprintf(pbuf, "\n   delta_t_LSF = %d  ", dt_lsf );
}

/**/
/* 0x54 */
static void rpt_1SV_bias (TSIPPKT *rpt)
{
	float
		clock_bias, freq_offset, time_of_fix;

	/* unload rptbuf */
	if (rpt_0x54 (rpt, &clock_bias, &freq_offset, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf (pbuf, "\nTime Fix   Clock Bias: %6.2f m  Freq Bias: %6.2f m/s%s",
		clock_bias, freq_offset, show_time (time_of_fix));
}

/* 0x55 */
static void rpt_io_opt (TSIPPKT *rpt)
{
	unsigned char
		pos_code, vel_code, time_code, aux_code;

	/* unload rptbuf */
	if (rpt_0x55 (rpt,
		&pos_code, &vel_code, &time_code, &aux_code)) {
		parsed = BADLEN_PARSE;
		return;
	}
	/* rptbuf unloaded */

	pbuf += sprintf(pbuf, "\nI/O Options: %2X %2X %2X %2X",
		pos_code, vel_code, time_code, aux_code);

	if (pos_code & 0x01) {
		pbuf += sprintf(pbuf, "\n    ECEF XYZ position output");
	}

	if (pos_code & 0x02) {
		pbuf += sprintf(pbuf, "\n    LLA position output");
	}

	pbuf += sprintf(pbuf, (pos_code & 0x04)?
		"\n    MSL altitude output (Geoid height) ":
		"\n    WGS-84 altitude output");

	pbuf += sprintf(pbuf, (pos_code & 0x08)?
		"\n    MSL altitude input":
      "\n    WGS-84 altitude input");

	pbuf += sprintf(pbuf, (pos_code & 0x10)?
		"\n    Double precision":
      "\n    Single precision");

	if (pos_code & 0x20) {
		pbuf += sprintf(pbuf, "\n    All Enabled Superpackets");
	}

	if (vel_code & 0x01) {
		pbuf += sprintf(pbuf, "\n    ECEF XYZ velocity output");
	}

	if (vel_code & 0x02) {
		pbuf += sprintf(pbuf, "\n    ENU velocity output");
	}

	pbuf += sprintf(pbuf, (time_code & 0x01)?
		  "\n    Time tags in UTC":
        "\n    Time tags in GPS time");

	if (time_code & 0x02) {
		pbuf += sprintf(pbuf, "\n    Fixes delayed to integer seconds");
	}

	if (time_code & 0x04) {
		pbuf += sprintf(pbuf, "\n    Fixes sent only on request");
	}

	if (time_code & 0x08) {
		pbuf += sprintf(pbuf, "\n    Synchronized measurements");
	}

	if (time_code & 0x10) {
		pbuf += sprintf(pbuf, "\n    Minimize measurement propagation");
	}

   pbuf += sprintf(pbuf, (time_code & 0x20) ?
		"\n    PPS output at all times" :
   	"\n    PPS output during fixes");

	if (aux_code & 0x01) {
		pbuf += sprintf(pbuf, "\n    Raw measurement output");
	}

	if (aux_code & 0x02) {
		pbuf += sprintf(pbuf, "\n    Code-phase smoothed before output");
	}

	if (aux_code & 0x04) {
		pbuf += sprintf(pbuf, "\n    Additional fix status");
	}

	pbuf += sprintf(pbuf, (aux_code & 0x08)?
   	"\n    Signal Strength Output as dBHz" :
   	"\n    Signal Strength Output as AMU");
}

/* 0x56 */
static void rpt_ENU_velocity (TSIPPKT *rpt)
{
	float
		vel_ENU[3], freq_offset, time_of_fix;

	/* unload rptbuf */
	if (rpt_0x56 (rpt, vel_ENU, &freq_offset, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nVel ENU: %11.3f  %11.3f  %11.3f  %12.3f%s",
		vel_ENU[0], vel_ENU[1], vel_ENU[2], freq_offset,
      show_time (time_of_fix));
}

/* 0x57 */
static void rpt_last_fix_info (TSIPPKT *rpt)
{
	unsigned char
		source_code, diag_code;
	short
		week_num;
	float
		time_of_fix;

	/* unload rptbuf */
	if (rpt_0x57 (rpt, &source_code, &diag_code, &week_num, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\n source code %d;   diag code: %2Xh",
   	source_code, diag_code);
	pbuf += sprintf(pbuf, "\n    Time of last fix:%s", show_time(time_of_fix));
	pbuf += sprintf(pbuf, "\n    Week of last fix: %d", week_num);
}

/* 0x58 */
static void rpt_GPS_system_data (TSIPPKT *rpt)
{
	unsigned char
   	iprn,
		op_code, data_type, sv_prn,
		data_length, data_packet[250];
	ALM_INFO
		*almanac;
	ALH_PARMS
		*almh;
	UTC_INFO
		*utc;
	ION_INFO
		*ionosphere;
	EPHEM_CLOCK
		*cdata;
	EPHEM_ORBIT
		*edata;
	NAV_INFO
		*nav_data;
	unsigned char
		curr_t_oa;
	unsigned short
		curr_wn_oa;
	static char
		*datname[] =
		{"", "", "Almanac Orbit",
		"Health Page & Ref Time", "Ionosphere", "UTC ",
		"Ephemeris"};

	/* unload rptbuf */
	if (rpt_0x58 (rpt, &op_code, &data_type, &sv_prn,
		&data_length, data_packet))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nSystem data [%d]:  %s  SV%02d",
		data_type, datname[data_type], sv_prn);
	switch (op_code)
	{
	case 1:
		pbuf += sprintf(pbuf, "  Acknowledgment");
		break;
	case 2:
		pbuf += sprintf(pbuf, "  length = %d bytes", data_length);
		switch (data_type) {
		case 2:
			/* Almanac */
			if (sv_prn == 0 || sv_prn > 32) {
				pbuf += sprintf(pbuf, "  Binary PRN invalid");
				return;
			}
			almanac = (ALM_INFO*)data_packet;
			pbuf += sprintf(pbuf, "\n   t_oa_raw = % -12d    SV_hlth  = % -12d  ",
         	almanac->t_oa_raw , almanac->SV_health );
			pbuf += sprintf(pbuf, "\n   e        = % -12g    t_oa     = % -12g  ",
         	almanac->e        , almanac->t_oa     );
			pbuf += sprintf(pbuf, "\n   i_0      = % -12g    OMEGADOT = % -12g  ",
         	almanac->i_0      , almanac->OMEGADOT );
			pbuf += sprintf(pbuf, "\n   sqrt_A   = % -12g    OMEGA_0  = % -12g  ",
         	almanac->sqrt_A   , almanac->OMEGA_0  );
			pbuf += sprintf(pbuf, "\n   omega    = % -12g    M_0      = % -12g  ",
         	almanac->omega    , almanac->M_0      );
			pbuf += sprintf(pbuf, "\n   a_f0     = % -12g    a_f1     = % -12g  ",
         	almanac->a_f0     , almanac->a_f1     );
			pbuf += sprintf(pbuf, "\n   Axis     = % -12g    n        = % -12g  ",
         	almanac->Axis     , almanac->n        );
			pbuf += sprintf(pbuf, "\n   OMEGA_n  = % -12g    ODOT_n   = % -12g  ",
         	almanac->OMEGA_n  , almanac->ODOT_n   );
			pbuf += sprintf(pbuf, "\n   t_zc     = % -12g    weeknum  = % -12d  ",
         	almanac->t_zc     , almanac->weeknum  );
			pbuf += sprintf(pbuf, "\n   wn_oa    = % -12d", almanac->wn_oa    );
			break;

		case 3:
			/* Almanac health page */
			almh = (ALH_PARMS*)data_packet;
			pbuf += sprintf(pbuf, "\n   t_oa = %d, wn_oa&0xFF = %d  ",
         	almh->t_oa, almh->WN_a);
			pbuf += sprintf(pbuf, "\nAlmanac health page:");
			for (iprn = 0; iprn < 32; iprn++) {
				if (!(iprn%5)) *pbuf++ = '\n';
				pbuf += sprintf(pbuf, "    SV%02d  %2X",
            	(iprn+1) , almh->SV_health[iprn]);
			}
			curr_t_oa = data_packet[34];
			curr_wn_oa = (unsigned short)((data_packet[35]<<8) + data_packet[36]);
			pbuf += sprintf(pbuf, "\n   current t_oa = %d, wn_oa = %d  ",
         	curr_t_oa, curr_wn_oa);
			break;

		case 4:
			/* Ionosphere */
			ionosphere = (ION_INFO*)data_packet;
			pbuf += sprintf(pbuf, "\n   alpha_0 = % -12g  alpha_1 = % -12g ",
	         ionosphere->alpha_0, ionosphere->alpha_1);
			pbuf += sprintf(pbuf, "\n   alpha_2 = % -12g  alpha_3 = % -12g ",
	         ionosphere->alpha_2, ionosphere->alpha_3);
			pbuf += sprintf(pbuf, "\n   beta_0  = % -12g  beta_1  = % -12g  ",
	         ionosphere->beta_0, ionosphere->beta_1);
			pbuf += sprintf(pbuf, "\n   beta_2  = % -12g  beta_3  = % -12g  ",
	         ionosphere->beta_2, ionosphere->beta_3);
			break;

		case 5:
			/* UTC */
			utc = (UTC_INFO*)data_packet;
			pbuf += sprintf(pbuf, "\n   A_0         = %g  ", utc->A_0);
			pbuf += sprintf(pbuf, "\n   A_1         = %g  ", utc->A_1);
			pbuf += sprintf(pbuf, "\n   delta_t_LS  = %d  ", utc->delta_t_LS);
			pbuf += sprintf(pbuf, "\n   t_ot        = %.0f  ", utc->t_ot );
			pbuf += sprintf(pbuf, "\n   WN_t        = %d  ", utc->WN_t );
			pbuf += sprintf(pbuf, "\n   WN_LSF      = %d  ", utc->WN_LSF );
			pbuf += sprintf(pbuf, "\n   DN          = %d  ", utc->DN );
			pbuf += sprintf(pbuf, "\n   delta_t_LSF = %d  ", utc->delta_t_LSF );
			break;

		case 6: /* Ephemeris */
			if (sv_prn == 0 || sv_prn > 32) {
				pbuf += sprintf(pbuf, "  Binary PRN invalid");
				return;
			}
			nav_data = (NAV_INFO*)data_packet;

			pbuf += sprintf(pbuf, "\n     SV_PRN = % -12d .  t_ephem = % -12g . ",
         	nav_data->sv_number , nav_data->t_ephem );
			cdata = &(nav_data->ephclk);
			pbuf += sprintf(pbuf,
         	"\n    weeknum = % -12d .   codeL2 = % -12d .  L2Pdata = % -12d",
         	cdata->weeknum , cdata->codeL2 , cdata->L2Pdata );
			pbuf += sprintf(pbuf,
         	"\n  SVacc_raw = % -12d .SV_health = % -12d .     IODC = % -12d",
         	cdata->SVacc_raw, cdata->SV_health, cdata->IODC );
			pbuf += sprintf(pbuf,
         	"\n       T_GD = % -12g .     t_oc = % -12g .     a_f2 = % -12g",
         	cdata->T_GD, cdata->t_oc, cdata->a_f2 );
			pbuf += sprintf(pbuf,
         	"\n       a_f1 = % -12g .     a_f0 = % -12g .    SVacc = % -12g",
         	cdata->a_f1, cdata->a_f0, cdata->SVacc );
			edata = &(nav_data->ephorb);
			pbuf += sprintf(pbuf,
	         "\n       IODE = % -12d .fit_intvl = % -12d .     C_rs = % -12g",
	         edata->IODE, edata->fit_interval, edata->C_rs );
			pbuf += sprintf(pbuf,
         	"\n    delta_n = % -12g .      M_0 = % -12g .     C_uc = % -12g",
         	edata->delta_n, edata->M_0, edata->C_uc );
			pbuf += sprintf(pbuf,
         	"\n        ecc = % -12g .     C_us = % -12g .   sqrt_A = % -12g",
	         edata->e, edata->C_us, edata->sqrt_A );
			pbuf += sprintf(pbuf,
         	"\n       t_oe = % -12g .     C_ic = % -12g .  OMEGA_0 = % -12g",
            edata->t_oe, edata->C_ic, edata->OMEGA_0 );
			pbuf += sprintf(pbuf,
	         "\n       C_is = % -12g .      i_0 = % -12g .     C_rc = % -12g",
	         edata->C_is, edata->i_0, edata->C_rc );
			pbuf += sprintf(pbuf,
	         "\n      omega = % -12g . OMEGADOT = % -12g .     IDOT = % -12g",
         	edata->omega, edata->OMEGADOT, edata->IDOT );
			pbuf += sprintf(pbuf,
   	      "\n       Axis = % -12g .        n = % -12g .    r1me2 = % -12g",
	         edata->Axis, edata->n, edata->r1me2 );
			pbuf += sprintf(pbuf,
      	   "\n    OMEGA_n = % -12g .   ODOT_n = % -12g",
	         edata->OMEGA_n, edata->ODOT_n );
			break;
		}
	}
}


/* 0x59: */
static void rpt_SVs_enabled (TSIPPKT *rpt)
{
	unsigned char
   	numsvs,
		code_type,
      status_code[32];
	short
		iprn;

	/* unload rptbuf */
	if (rpt_0x59 (rpt, &code_type, status_code))
   {
		parsed = BADLEN_PARSE;
		return;
	}
   switch (code_type)
   {
   case 3: pbuf += sprintf(pbuf, "\nSVs Disabled:\n"); break;
   case 6: pbuf += sprintf(pbuf, "\nSVs with Health Ignored:\n"); break;
   default: return;
   }
   numsvs = 0;
	for (iprn=0; iprn<32; iprn++)
   {
     	if (status_code[iprn])
      {
	   	pbuf += sprintf(pbuf, " %02d", iprn+1);
   	   numsvs++;
      }
   }
   if (numsvs == 0) pbuf += sprintf(pbuf, "None");
}


/* 0x5A */
static void rpt_raw_msmt (TSIPPKT *rpt)
{
	unsigned char
		sv_prn;
	float
		sample_length, signal_level, code_phase, Doppler;
	double
		time_of_fix;

	/* unload rptbuf */
	if (rpt_0x5A (rpt, &sv_prn, &sample_length, &signal_level,
		&code_phase, &Doppler, &time_of_fix))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\n   %02d %5.0f %7.1f %10.2f %10.2f %12.3f %s",
		sv_prn, sample_length, signal_level, code_phase, Doppler, time_of_fix,
		show_time ((float)time_of_fix));
}

/* 0x5B */
static void rpt_SV_ephemeris_status (TSIPPKT *rpt)
{
	unsigned char
		sv_prn, sv_health, sv_iode, fit_interval_flag;
	float
		time_of_collection, time_of_eph, sv_accy;

	/* unload rptbuf */
	if (rpt_0x5B (rpt, &sv_prn, &sv_health, &sv_iode, &fit_interval_flag,
		&time_of_collection, &time_of_eph, &sv_accy))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\n  SV%02d  %s   %2Xh     %2Xh ",
   	sv_prn, show_time (time_of_collection), sv_health, sv_iode);
	/* note: cannot use show_time twice in same call */
	pbuf += sprintf(pbuf, "%s   %1d   %4.1f",
      show_time (time_of_eph), fit_interval_flag, sv_accy);
}

/* 0x5C */
static void rpt_SV_tracking_status (TSIPPKT *rpt)
{
	unsigned char
		sv_prn, chan, slot, acq_flag, eph_flag,
		old_msmt_flag, integer_msec_flag, bad_data_flag,
		data_collect_flag;
	float
		signal_level, time_of_last_msmt,
		elev, azim;

	/* unload rptbuf */
	if (rpt_0x5C (rpt,
		&sv_prn, &slot, &chan, &acq_flag, &eph_flag,
		&signal_level, &time_of_last_msmt, &elev, &azim,
		&old_msmt_flag, &integer_msec_flag, &bad_data_flag,
		&data_collect_flag))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf,
"\n SV%2d  %1d   %1d   %1d   %4.1f  %s  %5.1f  %5.1f",
		sv_prn, chan,
      acq_flag, eph_flag, signal_level,
      show_time(time_of_last_msmt),
		elev*R2D, azim*R2D);
}

/**/
/* 0x6D */
static void rpt_allSV_selection (TSIPPKT *rpt)
{
	unsigned char
		manual_mode, nsvs, sv_prn[8], ndim;
	short
		islot;
	float
		pdop, hdop, vdop, tdop;

	/* unload rptbuf */
	if (rpt_0x6D (rpt,
		&manual_mode, &nsvs, &ndim, sv_prn,
		&pdop, &hdop, &vdop, &tdop))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	switch (ndim)
   {
   case 0:
		pbuf += sprintf(pbuf, "\nMode: Searching, %d-SV:", nsvs);
      break;
   case 1:
		pbuf += sprintf(pbuf, "\nMode: One-SV Timing:");
      break;
   case 3: case 4:
		pbuf += sprintf(pbuf, "\nMode: %c-%dD, %d-SV:",
   			manual_mode ? 'M' : 'A', ndim - 1,  nsvs);
      break;
	case 5:
		pbuf += sprintf(pbuf, "\nMode: Timing, %d-SV:", nsvs);
      break;
   default:
		pbuf += sprintf(pbuf, "\nMode: Unknown = %d:", ndim);
      break;
   }

	for (islot = 0; islot < nsvs; islot++)
   {
		if (sv_prn[islot]) pbuf += sprintf(pbuf, " %02d", sv_prn[islot]);
	}
   if (ndim == 3 || ndim == 4)
   {
		pbuf += sprintf(pbuf, ";  DOPs: P %.1f H %.1f V %.1f T %.1f",
			pdop, hdop, vdop, tdop);
   }
}

/**/
/* 0x82 */
static void rpt_DGPS_position_mode (TSIPPKT *rpt)
{
	unsigned char
		diff_mode;

	/* unload rptbuf */
	if (rpt_0x82 (rpt, &diff_mode)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nFix is%s DGPS-corrected (%s mode)  (%d)",
   	(diff_mode&1) ? "" : " not",
   	(diff_mode&2) ? "auto" : "manual",
      diff_mode);
}

/* 0x83 */
static void rpt_double_ECEF_position (TSIPPKT *rpt)
{

	double
		ECEF_pos[3], clock_bias;
	float
		time_of_fix;

	/* unload rptbuf */
	if (rpt_0x83 (rpt, ECEF_pos, &clock_bias, &time_of_fix))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nDXYZ:%12.2f  %13.2f  %13.2f %12.2f%s",
		ECEF_pos[0], ECEF_pos[1], ECEF_pos[2], clock_bias,
		show_time(time_of_fix));
}

/* 0x84 */
static void rpt_double_lla_position (TSIPPKT *rpt)
{
	short
		lat_deg, lon_deg;
	double
		lat, lon, lat_min, lon_min,
		alt, clock_bias;
	float
		time_of_fix;
	unsigned char
		north_south, east_west;

	/* unload rptbuf */
	if (rpt_0x84 (rpt,
		&lat, &lon, &alt, &clock_bias, &time_of_fix))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	lat *= R2D;
	lon *= R2D;
	if (lat < 0.0) {
		north_south = 'S';
		lat = -lat;
	} else {
		north_south = 'N';
	}
	lat_deg = (short)lat;
	lat_min = (lat - lat_deg) * 60.0;

	if (lon < 0.0) {
		east_west = 'W';
		lon = -lon;
	} else {
		east_west = 'E';
	}
	lon_deg = (short)lon;
	lon_min = (lon - lon_deg) * 60.0;
	pbuf += sprintf(pbuf, "\nDLLA: %2d:%08.5f %c; %3d:%08.5f %c; %10.2f %12.2f%s",
		lat_deg, lat_min, north_south,
		lon_deg, lon_min, east_west,
		alt, clock_bias,
		show_time(time_of_fix));
}

/* 0xBB */
static void rpt_complete_rcvr_config (TSIPPKT *rpt)
{
	TSIP_RCVR_CFG TsipxBB ;
	/* unload rptbuf */
	if (rpt_Paly0xBB (rpt, &TsipxBB))
	{
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\n   operating mode:      %s",
		NavModeText0xBB[TsipxBB.operating_mode]);
	pbuf += sprintf(pbuf, "\n   dynamics:            %s",
		dyn_text[TsipxBB.dyn_code]);
	pbuf += sprintf(pbuf, "\n   elev angle mask:     %g deg",
		TsipxBB.elev_mask * R2D);
	pbuf += sprintf(pbuf, "\n   SNR mask:            %g AMU",
		TsipxBB.cno_mask);
	pbuf += sprintf(pbuf, "\n   DOP mask:            %g",
		TsipxBB.dop_mask);
	pbuf += sprintf(pbuf, "\n   DOP switch:          %g",
		TsipxBB.dop_switch);
	return ;
}

/* 0xBC */
static void rpt_rcvr_serial_port_config (TSIPPKT *rpt)
{
	unsigned char
		port_num, in_baud, out_baud, data_bits, parity, stop_bits, flow_control,
		protocols_in, protocols_out, reserved;
	unsigned char known;

	/* unload rptbuf */
	if (rpt_0xBC (rpt, &port_num, &in_baud, &out_baud, &data_bits, &parity,
			&stop_bits, &flow_control, &protocols_in, &protocols_out, &reserved)) {
		parsed = BADLEN_PARSE;
		return;
	}
	/* rptbuf unloaded */

	pbuf += sprintf(pbuf, "\n   RECEIVER serial port %s config:",
		rcvr_port_text[port_num]);

	pbuf += sprintf(pbuf, "\n             I/O Baud %s/%s, %d - %s - %d",
		st_baud_text_app[in_baud],
		st_baud_text_app[out_baud],
		data_bits+5,
		parity_text[parity],
		stop_bits=1);
	pbuf += sprintf(pbuf, "\n             Input protocols: ");
	known = FALSE;
	if (protocols_in&TSIP)
   {
		pbuf += sprintf(pbuf, "%s ", protocols_in_text[1]);
		known = TRUE;
	}
	if (known == FALSE) pbuf += sprintf(pbuf, "No known");

	pbuf += sprintf(pbuf, "\n             Output protocols: ");
	known = FALSE;
	if (protocols_out&TSIP)
   {
		pbuf += sprintf(pbuf, "%s ", protocols_out_text[1]);
		known = TRUE;
	}
	if (protocols_out&NMEA)
   {
		pbuf += sprintf(pbuf, "%s ", protocols_out_text[2]);
		known = TRUE;
	}
	if (known == FALSE) pbuf += sprintf(pbuf, "No known");
	reserved = reserved;

 }

/* 0x8F */
/* 8F0B */
static void rpt_8F0B(TSIPPKT *rpt)
{
	const char
   	*oprtng_dim[7] = {
      	"horizontal (2-D)",
         "full position (3-D)",
         "single satellite (0-D)",
         "automatic",
         "N/A",
         "N/A",
         "overdetermined clock"};
   char
   	sv_id[8];
   unsigned char
   	month,
      date,
      dim_mode,
      north_south,
      east_west;
   unsigned short
   	event;
   short
   	utc_offset,
      year,
      index;
	short
   	lat_deg,
      lon_deg;
   float
   	bias_unc,
      dr_unc;
   double
   	tow,
      bias,
      drift,
      lat,
      lon,
      alt,
      lat_min,
      lon_min;
   int
   	numfix,
      numnotfix;

	if (rpt_0x8F0B(rpt,
   	&event,
      &tow,
      &date,
      &month,
      &year,
      &dim_mode,
      &utc_offset,
      &bias,
      &drift,
      &bias_unc,
      &dr_unc,
      &lat,
      &lon,
      &alt,
      sv_id))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	if (event == 0)
   {
   	pbuf += sprintf(pbuf, "\nNew partial+full meas");
	}
   else
   {
		pbuf += sprintf(pbuf, "\nEvent count: %5d", event);
   }

	pbuf += sprintf(pbuf, "\nGPS time  : %s %2d/%2d/%2d (DMY)",
   	show_time(tow), date, month, year);
	pbuf += sprintf(pbuf, "\nMode      : %s", oprtng_dim[dim_mode]);
	pbuf += sprintf(pbuf, "\nUTC offset: %2d", utc_offset);
	pbuf += sprintf(pbuf, "\nClock Bias: %6.2f m", bias);
	pbuf += sprintf(pbuf, "\nFreq bias : %6.2f m/s", drift);
	pbuf += sprintf(pbuf, "\nBias unc  : %6.2f m", bias_unc);
	pbuf += sprintf(pbuf, "\nFreq unc  : %6.2f m/s", dr_unc);

	lat *= R2D; /* convert from radians to degrees */
	lon *= R2D;
	if (lat < 0.0)
   {
		north_south = 'S';
		lat = -lat;
	}
   else
   {
		north_south = 'N';
	}

	lat_deg = (short)lat;
	lat_min = (lat - lat_deg) * 60.0;
	if (lon < 0.0)
   {
		east_west = 'W';
		lon = -lon;
	}
   else
   {
		east_west = 'E';
	}

	lon_deg = (short)lon;
	lon_min = (lon - lon_deg) * 60.0;
	pbuf += sprintf(pbuf, "\nPosition  :");
	pbuf += sprintf(pbuf, " %4d %6.3f %c", lat_deg, lat_min, north_south);
	pbuf += sprintf(pbuf, " %5d %6.3f %c", lon_deg, lon_min, east_west);
	pbuf += sprintf(pbuf, " %10.2f", alt);

   numfix = numnotfix = 0;
	for (index=0; index<8; index++)
   {
		if (sv_id[index] < 0) numnotfix++;
		if (sv_id[index] > 0) numfix++;
   }
   if (numfix > 0)
   {
		pbuf += sprintf(pbuf, "\nSVs used in fix  : ");
		for (index=0; index<8; index++)
	   {
			if (sv_id[index] > 0)
      	{
      		pbuf += sprintf(pbuf, "%2d ", sv_id[index]);
	      }
   	}
   }
   if (numnotfix > 0)
   {
		pbuf += sprintf(pbuf, "\nOther SVs tracked: ");
		for (index=0; index<8; index++)
	   {
			if (sv_id[index] < 0)
      	{
      		pbuf += sprintf(pbuf, "%2d ", sv_id[index]);
	      }
   	}
   }
}

/* 0x8F14 */
static void rpt_8F14 (TSIPPKT *rpt)
/* Datum parameters */
{
	double
		datum_coeffs[5];
	short
		datum_idx;

	/* unload rptbuf */
	if (rpt_0x8F14 (rpt, &datum_idx, datum_coeffs))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	if (datum_idx == -1)
   {
   	pbuf += sprintf(pbuf, "\nUser-Entered Datum:");
		pbuf += sprintf(pbuf, "\n   dx        = %6.1f", datum_coeffs[0]);
		pbuf += sprintf(pbuf, "\n   dy        = %6.1f", datum_coeffs[1]);
		pbuf += sprintf(pbuf, "\n   dz        = %6.1f", datum_coeffs[2]);
		pbuf += sprintf(pbuf, "\n   a-axis    = %10.3f", datum_coeffs[3]);
		pbuf += sprintf(pbuf, "\n   e-squared = %16.14f", datum_coeffs[4]);
   }
   else if (datum_idx == 0)
   {
   	pbuf += sprintf(pbuf, "\nWGS-84 datum, Index 0 ");
   }
   else
   {
   	pbuf += sprintf(pbuf, "\nStandard Datum, Index %3d ", datum_idx);
   }
}

/* 0x8F15 */
static void rpt_8F15 (TSIPPKT *rpt)
/* Datum parameters */
{
	double
		datum_coeffs[5];
	short
		datum_idx;

	/* unload rptbuf */
	if (rpt_0x8F15 (rpt, &datum_idx, datum_coeffs)) {
		parsed = BADLEN_PARSE;
		return;
	}

	if (datum_idx == -1)
   {
   	pbuf += sprintf(pbuf, "\nUser-Entered Datum:");
		pbuf += sprintf(pbuf, "\n   dx        = %6.1f", datum_coeffs[0]);
		pbuf += sprintf(pbuf, "\n   dy        = %6.1f", datum_coeffs[1]);
		pbuf += sprintf(pbuf, "\n   dz        = %6.1f", datum_coeffs[2]);
		pbuf += sprintf(pbuf, "\n   a-axis    = %10.3f", datum_coeffs[3]);
		pbuf += sprintf(pbuf, "\n   e-squared = %16.14f", datum_coeffs[4]);
   }
   else if (datum_idx == 0)
   {
   	pbuf += sprintf(pbuf, "\nWGS-84 datum, Index 0 ");
   }
   else
   {
   	pbuf += sprintf(pbuf, "\nStandard Datum, Index %3d ", datum_idx);
   }
}

/* 0x8F20 */
#define INFO_DGPS       0x02
#define INFO_2D         0x04
#define INFO_ALTSET     0x08
#define INFO_FILTERED   0x10
static void rpt_8F20 (TSIPPKT *rpt)
{
	unsigned char
		info, nsvs, sv_prn[32];
	short
		week_num, datum_index, sv_IODC[32];
	double
		lat, lon, alt, time_of_fix;
	double
		londeg, latdeg, vel[3];
	short
		isv;
   char
   	datum_string[20];

	/* unload rptbuf */
	if (rpt_0x8F20 (rpt,
		&info, &lat, &lon, &alt, vel,
		&time_of_fix,
		&week_num, &nsvs, sv_prn, sv_IODC, &datum_index))
	{
		parsed = BADLEN_PARSE;
		return;
	}
	pbuf += sprintf(pbuf,
   	"\nFix at: %04d:%3s:%02d:%02d:%06.3f GPS (=UTC+%2ds)  FixType: %s%s%s",
   	week_num,
		dayname[(short)(time_of_fix/86400.0)],
		(short)fmod(time_of_fix/3600., 24.),
		(short)fmod(time_of_fix/60., 60.),
		fmod(time_of_fix, 60.),
      (char)rpt->buf[29],		/* UTC offset */
		(info & INFO_DGPS)?"Diff":"",
		(info & INFO_2D)?"2D":"3D",
		(info & INFO_FILTERED)?"-Filtrd":"");

   if (datum_index > 0)
   {
		sprintf(datum_string, "Datum%3d", datum_index);
   }
   else if (datum_index)
   {
		sprintf(datum_string, "Unknown ");
   }
   else
   {
		sprintf(datum_string, "WGS-84");
   }

	/* convert from radians to degrees */
	latdeg = R2D * fabs(lat);
	londeg = R2D * fabs(lon);
	pbuf += sprintf(pbuf,
   	"\n   Pos: %4d:%09.6f %c %5d:%09.6f %c %10.2f m HAE (%s)",
		(short)latdeg, fmod (latdeg, 1.)*60.0,
		(lat<0.0)?'S':'N',
		(short)londeg, fmod (londeg, 1.)*60.0,
		(lon<0.0)?'W':'E',
		alt,
      datum_string);
	pbuf += sprintf(pbuf,
   	"\n   Vel:    %9.3f E       %9.3f N      %9.3f U   (m/sec)",
		vel[0], vel[1], vel[2]);

	pbuf += sprintf(pbuf,
   	"\n   SVs: ");
	for (isv = 0; isv < nsvs; isv++) {
		pbuf += sprintf(pbuf, " %02d", sv_prn[isv]);
	}
	pbuf += sprintf(pbuf, "     (IODEs:");
	for (isv = 0; isv < nsvs; isv++) {
		pbuf += sprintf(pbuf, " %02X", sv_IODC[isv]&0xFF);
	}
	pbuf += sprintf(pbuf, ")");
}

/* 0x8F41 */
static void rpt_8F41(TSIPPKT *rpt)
{
	unsigned char
   	bSearchRange,
		bBoardOptions,
		bBuildYear,
		bBuildMonth,
		bBuildDay,
		bBuildHour;
	float
   	fOscOffset;
	unsigned short
   	iTestCodeId;
	unsigned long
		iiSerialNumber;

   if (!rpt_0x8F41(rpt,
		&bSearchRange,
		&bBoardOptions,
		&iiSerialNumber,
		&bBuildYear,
		&bBuildMonth,
		&bBuildDay,
		&bBuildHour,
		&fOscOffset,
		&iTestCodeId))
   {
		parsed = BADLEN_PARSE;
      return;
   }

   pbuf += sprintf(pbuf, "\n  search range:          %d",
   	bSearchRange);
   pbuf += sprintf(pbuf, "\n  board options:         %d",
      bBoardOptions);
   pbuf += sprintf(pbuf, "\n  board serial #:        %ld",
      iiSerialNumber);
   pbuf += sprintf(pbuf, "\n  build date/hour:       %02d/%02d/%02d %02d:00",
   	bBuildDay, bBuildMonth, bBuildYear, bBuildHour);
   pbuf += sprintf(pbuf, "\n  osc offset:            %.3f PPM (%.0f Hz)",
   	fOscOffset/1575.42, fOscOffset);
   pbuf += sprintf(pbuf, "\n  test code:             %d",
   	iTestCodeId);
}

/* 0x8F42 */
static void rpt_8F42(TSIPPKT *rpt)
{
	unsigned char
   	bProdOptionsPre,
      bProdNumberExt;
	unsigned short
   	iCaseSerialNumberPre,
      iPremiumOptions,
      iMachineID,
      iKey;
	unsigned long
   	iiCaseSerialNumber,
		iiProdNumber;

   if (!rpt_0x8F42(rpt,
		&bProdOptionsPre,
		&bProdNumberExt,
		&iCaseSerialNumberPre,
		&iiCaseSerialNumber,
		&iiProdNumber,
		&iPremiumOptions,
		&iMachineID,
		&iKey))
   {
		parsed = BADLEN_PARSE;
      return;
   }

	pbuf += sprintf(pbuf, "\nProduct ID 8F42");
   pbuf += sprintf(pbuf, "\n   extension:            %d", bProdNumberExt);
   pbuf += sprintf(pbuf, "\n   case serial # prefix: %d", iCaseSerialNumberPre);
   pbuf += sprintf(pbuf, "\n   case serial #:        %ld", iiCaseSerialNumber);
   pbuf += sprintf(pbuf, "\n   prod. #:              %ld", iiProdNumber);
	pbuf += sprintf(pbuf, "\n   premium options:      %Xh", iPremiumOptions);
   pbuf += sprintf(pbuf, "\n   machine ID:           %d", iMachineID);
   pbuf += sprintf(pbuf, "\n   key:                  %Xh", iKey);
}

/* 0x8F45 */
static void rpt_8F45(TSIPPKT *rpt)
{
   unsigned char bSegMask;

   if (!rpt_0x8F45(rpt,
   	&bSegMask))
   {
		parsed = BADLEN_PARSE;
		return;
	}
	pbuf += sprintf(pbuf, "\nCleared Segment Mask: %Xh", bSegMask);
}

static void rpt_8F4A(TSIPPKT *rpt)
/* Stinger PPS def */
{
	unsigned char
   	pps_enabled,
      pps_timebase,
      pps_polarity;
   float
   	bias_unc_threshold;
   double
   	pps_offset;

  	if (rpt_0x8F4A_16 (rpt,
   	&pps_enabled,
      &pps_timebase,
      &pps_polarity,
      &pps_offset,
      &bias_unc_threshold))
   {
   	parsed = BADLEN_PARSE;
	   return;
   }

	pbuf += sprintf(pbuf, "\nPPS is         %s",	pps_enabled?"enabled":"disabled");
   pbuf += sprintf(pbuf, "\n   timebase:   %s", PPSTimeBaseText[pps_timebase]);
   pbuf += sprintf(pbuf, "\n   polarity:   %s", PPSPolarityText[pps_polarity]);
   pbuf += sprintf(pbuf, "\n   offset:     %.1f ns, ", pps_offset*1.e9);
   pbuf += sprintf(pbuf, "\n   biasunc:    %.1f ns", bias_unc_threshold/GPS_C*1.e9);
}

static void rpt_8F4B(TSIPPKT *rpt)
/* fast-SA decorrolation time for self-survey */
{
	unsigned long
   	decorr_max;

   if (rpt_0x8F4B(rpt, &decorr_max))
   {
		parsed = BADLEN_PARSE;
      return;
   }

   pbuf += sprintf(pbuf,
   	"\nMax # of position fixes for self-survey : %ld",
      decorr_max);
}

static void rpt_8F4D(TSIPPKT *rpt)
{
	static char
   	*linestart;
	unsigned long
   	OutputMask;
   static unsigned long
   	MaskBit[] = {
      	0x00000001, 0x00000002, 0x00000004, 0x00000008, 0x00000010, 0x00000020,
      	0x00000100L, 0x00000800L, 0x00001000L,
         0x40000000L, 0x80000000L};
   int
   	ichoice,
   	numchoices;

   if (rpt_0x8F4D(rpt, &OutputMask))
   {
		parsed = BADLEN_PARSE;
      return;
   }

   pbuf += sprintf(pbuf, "\nAuto-Report Mask: %02X %02X %02X %02X",
   	(unsigned char)(OutputMask>>24),
   	(unsigned char)(OutputMask>>16),
   	(unsigned char)(OutputMask>>8),
   	(unsigned char)OutputMask);

   numchoices = sizeof(MaskText)/sizeof(char*);
   pbuf += sprintf(pbuf, "\nAuto-Reports scheduled for Output:");
   linestart = pbuf;
   for (ichoice=0; ichoice<numchoices; ichoice++)
   {
   	if (OutputMask&MaskBit[ichoice])
      {
	     	pbuf += sprintf(pbuf, "%s %s",
   	   	(pbuf==linestart)?"\n     ":",",
      	   MaskText[ichoice]);
			if (pbuf-linestart > 60) linestart = pbuf;
      }
   }

   pbuf += sprintf(pbuf, "\nAuto-Reports NOT scheduled for Output:");
   linestart = pbuf;
   for (ichoice=0; ichoice<numchoices; ichoice++)
   {
   	if (OutputMask&MaskBit[ichoice]) continue;
	     	pbuf += sprintf(pbuf, "%s %s",
   	   	(pbuf==linestart)?"\n     ":",",
         MaskText[ichoice]);
		if (pbuf-linestart > 60) linestart = pbuf;
   }
}

static void rpt_8FA5(TSIPPKT *rpt)
{
	unsigned char
   	spktmask[4];

   if (rpt_0x8FA5(rpt, spktmask))
   {
		parsed = BADLEN_PARSE;
      return;
   }

   pbuf += sprintf(pbuf, "\nSuperpacket auto-output mask: %02X %02X %02X %02X",
   	spktmask[0], spktmask[1], spktmask[2], spktmask[3]);

   if (spktmask[0]&0x01) pbuf+= sprintf (pbuf, "\n    PPS   8F-0B");
   if (spktmask[0]&0x02) pbuf+= sprintf (pbuf, "\n    Event 8F-0B");
   if (spktmask[0]&0x10) pbuf+= sprintf (pbuf, "\n    PPS   8F-AD");
   if (spktmask[0]&0x20) pbuf+= sprintf (pbuf, "\n    Event 8F-AD");
   if (spktmask[2]&0x01) pbuf+= sprintf (pbuf, "\n    ppos Fix 8F-20");
}

static void rpt_8FAD (TSIPPKT *rpt)
{
   unsigned short
    	Count,
    	Year;
   double
    	FracSec;
   unsigned char
    	Hour,
    	Minute,
    	Second,
    	Day,
    	Month,
    	Status,
    	Flags;
	static char* Status8FADText[] = {
      "CODE_DOING_FIXES",
      "CODE_GOOD_1_SV",
      "CODE_APPX_1SV",
      "CODE_NEED_TIME",
      "CODE_NEED_INITIALIZATION",
      "CODE_PDOP_HIGH",
      "CODE_BAD_1SV",
      "CODE_0SVS",
      "CODE_1SV",
      "CODE_2SVS",
      "CODE_3SVS",
      "CODE_NO_INTEGRITY",
      "CODE_DCORR_GEN",
      "CODE_OVERDET_CLK",
      "Invalid Status"},
    	*LeapStatusText[] = {
    	" UTC Avail", " ", " ", " ",
      " Scheduled", " Pending", " Warning", " In Progress"};
    int i;

	if (rpt_0x8FAD (rpt,
    	&Count,
    	&FracSec,
    	&Hour,
    	&Minute,
    	&Second,
    	&Day,
    	&Month,
    	&Year,
    	&Status,
    	&Flags))
   {
		parsed = BADLEN_PARSE;
		return;
   }

	pbuf += sprintf(pbuf,    "\n8FAD   Count: %d   Status: %s",
   	Count, Status8FADText[Status]);

  	pbuf += sprintf(pbuf, "\n   Leap Flags:", Flags);
   if (Flags)
   {
   	for (i=0; i<8; i++)
      {
      	if (Flags&(1<<i)) pbuf += sprintf(pbuf, LeapStatusText[i]);
      }
   }
   else
   {
   	pbuf += sprintf(pbuf, "  UTC info not available");
   }

	pbuf += sprintf(pbuf,     "\n      %02d/%02d/%04d (DMY)  %02d:%02d:%02d.%09d UTC",
   		Day, Month, Year, Hour, Minute, Second, (long)(FracSec*1.e9));
}


int print_msg_table_header (int rptcode, char *HdrStr, int force)
{
	/* force header is to help auto-output function */
	/* last_rptcode is to determine whether to print a header */
	/* for the first occurence of a series of reports */
	static int
		last_rptcode = 0;
   int
   	numchars;

   numchars = 0;
	if (force || rptcode!=last_rptcode)
   {
		/* supply a header in console output */
   	switch (rptcode)
		{
		case 0x5A:
			numchars = sprintf(HdrStr, "\nRaw Measurement Data");
			numchars += sprintf(HdrStr+numchars,
      		"\n   SV  Sample   SNR  Code Phase   Doppler    Seconds     Time of Meas");
			break;

		case 0x5B:
			numchars = sprintf(HdrStr, "\nEphemeris Status");
			numchars += sprintf(HdrStr+numchars,
				"\n    SV     Time collected     Health  IODE        t oe         Fit   URA");
			break;

		case 0x5C:
			numchars = sprintf(HdrStr, "\nTracking Info");
			numchars += sprintf(HdrStr+numchars,
   	   	"\n   SV  C Acq Eph   SNR     Time of Meas       Elev  Azim   ");
			break;

      }
	}
	last_rptcode = rptcode;
   return (short)numchars;
}

static void unknown_rpt (TSIPPKT *rpt)
{
	int i;

	/* app-specific rpt packets */
	if (parsed == BADLEN_PARSE)
   {
		pbuf += sprintf(pbuf, "\nTSIP report packet ID %2Xh, length %d: Bad length",
      	rpt->code, rpt->len);
   }
	if (parsed == BADID_PARSE)
   {
		pbuf += sprintf(pbuf,
      	"\nTSIP report packet ID %2Xh, length %d: translation not supported",
   		rpt->code, rpt->len);
   }

	if (parsed == BADDATA_PARSE)
   {
		pbuf += sprintf(pbuf,
      	"\nTSIP report packet ID %2Xh, length %d: data content incorrect",
   		rpt->code, rpt->len);
   }

	for (i = 0; i < rpt->len; i++) {
		if ((i % 20) == 0) *pbuf++ = '\n';
		pbuf += sprintf(pbuf, " %02X", rpt->buf[i]);
	}
}
/**/
/*
** main subroutine, called from ProcessInputBytesWhileWaitingForKBHit()
*/
void TranslateTSIPReportToText (TSIPPKT *rpt, char *TextOutputBuffer)
{

	/* pbuf is the pointer to the current location of the text output */
	pbuf = TextOutputBuffer;

   /* keep track of whether the message has been successfully parsed */
	parsed = GOOD_PARSE;

	/* print a header if this is the first of a series of messages */
	pbuf += print_msg_table_header (rpt->code, pbuf, FALSE);

   /* process incoming TSIP report according to code */
	switch (rpt->code)
   {
	case 0x3D: rpt_chan_A_config (rpt); break;
	case 0x40: rpt_almanac_data_page (rpt); break;
	case 0x41: rpt_GPS_time (rpt); break;
	case 0x42: rpt_single_ECEF_position (rpt); break;
	case 0x43: rpt_single_ECEF_velocity (rpt); break;
	case 0x45: rpt_SW_version (rpt); break;
	case 0x46: rpt_rcvr_health (rpt); break;
	case 0x47: rpt_SNR_all_SVs (rpt); break;
	case 0x48: rpt_GPS_system_message (rpt); break;
	case 0x49: rpt_almanac_health_page (rpt); break;
	case 0x4A: switch (rpt->len) {
   	/*
      ** special case (=slip-up) in the TSIP protocol;
      ** parsing method depends on length
      */
   	case 20: rpt_single_lla_position (rpt); break;
      case  9: rpt_ref_alt (rpt); break;
		} break;
	case 0x4B: rpt_rcvr_id_and_status (rpt);break;
	case 0x4C: rpt_operating_parameters (rpt); break;
	case 0x4D: rpt_oscillator_offset (rpt); break;
	case 0x4E: rpt_GPS_time_set_response (rpt); break;
	case 0x4F: rpt_UTC_offset (rpt); break;
   case 0x54: rpt_1SV_bias (rpt); break;
	case 0x55: rpt_io_opt (rpt); break;
	case 0x56: rpt_ENU_velocity (rpt); break;
	case 0x57: rpt_last_fix_info (rpt); break;
	case 0x58: rpt_GPS_system_data (rpt); break;
	case 0x59: rpt_SVs_enabled (rpt); break;
	case 0x5A: rpt_raw_msmt (rpt); break;
	case 0x5B: rpt_SV_ephemeris_status (rpt); break;
	case 0x5C: rpt_SV_tracking_status (rpt); break;
	case 0x6D: rpt_allSV_selection (rpt); break;
	case 0x82: rpt_DGPS_position_mode (rpt); break;
	case 0x83: rpt_double_ECEF_position (rpt); break;
	case 0x84: rpt_double_lla_position (rpt); break;
	case 0xBB: rpt_complete_rcvr_config (rpt); break;
	case 0xBC: rpt_rcvr_serial_port_config (rpt); break;

	case 0x8F: switch (rpt->buf[0])
   	{
      /* superpackets; parsed according to subcodes */
      case 0x0B: rpt_8F0B(rpt); break;
      case 0x14: rpt_8F14(rpt); break;
      case 0x15: rpt_8F15(rpt); break;
		case 0x20: rpt_8F20(rpt); break;
      case 0x41: rpt_8F41(rpt); break;
      case 0x42: rpt_8F42(rpt); break;
      case 0x45: rpt_8F45(rpt); break;
      case 0x4A: rpt_8F4A(rpt); break;
      case 0x4B: rpt_8F4B(rpt); break;
      case 0x4D: rpt_8F4D(rpt); break;
      case 0xA5: rpt_8FA5(rpt); break;
 	   case 0xAD: rpt_8FAD(rpt); break;
		default: parsed = BADID_PARSE; break;
		}
		break;

	default: parsed = BADID_PARSE; break;
	}

	if (parsed != GOOD_PARSE)
	{
	   /*
   	**The message has TSIP structure (DLEs, etc.)
	   ** but could not be parsed by above routines
   	*/
		unknown_rpt (rpt);
	}

   /* close TextOutputBuffer */
   pbuf = '\0';
}


